@@ -1,5 +1,3 @@
-import uuid
-
 COLUMN_INDEX_TYPE_NONE = "none"
 COLUMN_INDEX_TYPE_INDEXED = "indexed"
 COLUMN_INDEX_TYPE_UNIQUE = "unique"
@@ -20,3 +18,6 @@ TRANSFORM_TYPE_PASSTHROUGH = "passthrough"
 TRANSFORM_TYPE_TOKENIZE_BY_REFERENCE = "tokenizebyreference"
 TRANSFORM_TYPE_TOKENIZE_BY_VALUE = "tokenizebyvalue"
 TRANSFORM_TYPE_TRANSFORM = "transform"
+
+MUTATOR_COLUMN_DEFAULT_VALUE = "UCDEF-7f55f479-3822-4976-a8a9-b789d5c6f152"
+MUTATOR_COLUMN_CURRENT_VALUE = "UCCUR-7f55f479-3822-4976-a8a9-b789d5c6f152"
